"""
## GAMSSOURCE: https://www.gams.com/latest/psoptlib_ml/libhtml/psoptlib_DED.html
## LICENSETYPE: Demo
## MODELTYPE: QCP


Dynamic Economic Load Dispatch

For more details please refer to Chapter 4 (Gcode4.1), of the following book:
Soroudi, Alireza. Power System Optimization Modeling in GAMS. Springer, 2017.
--------------------------------------------------------------------------------
Model type: QCP
--------------------------------------------------------------------------------
Contributed by
Dr. Alireza Soroudi
IEEE Senior Member
Email: alireza.soroudi@gmail.com
We do request that publications derived from the use of the developed GAMS code
explicitly acknowledge that fact by citing
Soroudi, Alireza. Power System Optimization Modeling in GAMS. Springer, 2017.
DOI: doi.org/10.1007/978-3-319-62350-4
"""

from __future__ import annotations

import numpy as np
import pandas as pd

from gamspy import Container, Equation, Model, Parameter, Set, Sum, Variable
from gamspy.math import sqr
import sys

def reformat_df(dataframe):
    return dataframe.reset_index().melt(
        id_vars="index", var_name="Category", value_name="Value"
    )


def read_csv_data(file_path):
    """
    Read timesteps and demand data from a CSV file.
    """
    df = pd.read_csv(file_path)
    timesteps = df['timestep'].tolist()
    demand_data = df['demand'].tolist()
    return timesteps, demand_data


def data_records(file_path):
    # gendata records table
    cols = ["a", "b", "c", "d", "e", "f", "Pmin", "Pmax", "RU0", "RD0"]
    inds = [f"g{i}" for i in range(1, 5)]
    data = [
        [0.12, 14.80, 89, 1.2, -5.0, 3.0, 28, 200, 40, 40],
        [0.17, 16.57, 83, 2.3, -4.24, 6.09, 20, 290, 30, 30],
        [0.15, 15.55, 100, 1.1, -2.15, 5.69, 30, 190, 30, 30],
        [0.19, 16.21, 70, 1.1, -3.99, 6.2, 20, 260, 50, 50],
    ]
    gen_recs = reformat_df(pd.DataFrame(data, columns=cols, index=inds))

    # Read demand data from CSV file
    timesteps, demand_recs = read_csv_data(file_path)

    return gen_recs, demand_recs, timesteps


def main():
    file_path = 'load_profile_full_year_test.csv'  # Update this path to your CSV file
    m = Container()

    # SETS #
    timesteps = data_records(file_path)[2]
    t = Set(
        m,
        name="t",
        records=timesteps,
        description="hours",
    )
    i = Set(
        m,
        name="i",
        records=[f"g{i}" for i in range(1, 5)],
        description="thermal units",
    )
    # PARAMETERS #
    demand_data = [(i1,i2*750) for i1, i2 in zip(data_records(file_path)[2], data_records(file_path)[1])]
    demand = Parameter(m, name="demand", domain=t, records=demand_data)
    demand_data_flex = [(i1,i2*750*0.05) for i1, i2 in zip(data_records(file_path)[2], data_records(file_path)[1])]    
    demand_flex = Parameter(m, name="demand_flex", domain=t, records=demand_data_flex)
    
    demand_met = Variable(m, name="demand_met", type="positive", domain=t, description="demand met by generators")
    demand_flex_met = Variable(m, name="demand_flex_met", type="positive", domain=t, description="demand met by generators")
    
    VOLL = Parameter(m, name= "VOLL", records=1000, description="Value of lost load")

    gendata = Parameter(
        m,
        name="gendata",
        domain=[i, "*"],
        records=data_records(file_path)[0],
        description="generator cost characteristics and limits",
    )

    # VARIABLES #
    p = Variable(
        m,
        name="p",
        type="free",
        domain=[i, t],
        description="power generated by thermal power plant",
    )
    EM = Variable(
        m, name="EM", type="free", description="emission calculation"
    )

    p.up[i, t] = gendata[i, "Pmax"]
    p.lo[i, t] = gendata[i, "Pmin"]

    # EQUATIONS #
    Genconst3 = Equation(m, name="Genconst3", type="regular", domain=[i, t])
    Genconst4 = Equation(m, name="Genconst4", type="regular", domain=[i, t])
    balance = Equation(m, name="balance", type="regular", domain=t)
    demand_lim = Equation(m, name="demand_lim", type="regular", domain = t)
    demand_flex_lim = Equation(m, name="demand_flex_lim", type="regular", domain = t)

    EMcalc = Equation(m, name="EMcalc", type="regular")
    
    demand_lim[t] = demand_met[t] <= demand[t]
    demand_flex_lim[t] = demand_flex_met[t] <= demand_flex[t]

    costThermal = Sum(
        [t, i],
        gendata[i, "a"] * sqr(p[i, t])
        + gendata[i, "b"] * p[i, t]
        + gendata[i, "c"],
    ) - Sum(t, VOLL * ( demand_met[t] + demand_flex_met[t] - demand_flex_met[t]*demand_flex_met[t]/(2*demand_flex[t])))

    Genconst3[i, t] = p[i, t + 1] - p[i, t] <= gendata[i, "RU0"]

    Genconst4[i, t] = p[i, t - 1] - p[i, t] <= gendata[i, "RD0"]

    balance[t] = Sum(i, p[i, t]) >= demand_met[t] + demand_flex_met[t]

    EMcalc[...] = (
        Sum(
            [t, i],
            gendata[i, "d"] * sqr(p[i, t])
            + gendata[i, "e"] * p[i, t]
            + gendata[i, "f"],
        )
        == EM
    )

    DEDcostbased = Model(
        m,
        name="DEDcostbased",
        equations=m.getEquations(),
        problem="qcp",
        sense="min",
        objective=costThermal,
    )
    for i in range(7):
        DEDcostbased.solve( solver="cplex", output=sys.stdout)



    # Export results to excel


if __name__ == "__main__":
    main()